@@ -2,22 +2,39 @@
  *
  * server rate limiting
  *
- * Copyright (c) 2005 by Rhyolite Software
+ * Copyright (c) 2005 by Rhyolite Software, LLC
+ *
+ * This agreement is not applicable to any entity which sells anti-spam
+ * solutions to others or provides an anti-spam solution as part of a
+ * security solution sold to other entities, or to a private network
+ * which employs the DCC or uses data provided by operation of the DCC
+ * but does not provide corresponding data to other users.
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
  * copyright notice and this permission notice appear in all copies.
  *
- * THE SOFTWARE IS PROVIDED "AS IS" AND RHYOLITE SOFTWARE DISCLAIMS ALL
+ * Parties not eligible to receive a license under this agreement can
+ * obtain a commercial license to use DCC and permission to use
+ * U.S. Patent 6,330,590 by contacting Commtouch at http://www.commtouch.com/
+ * or by email to nospam@commtouch.com.
+ *
+ * A commercial license would be for Distributed Checksum and Reputation
+ * Clearinghouse software.  That software includes additional features.  This
+ * free license for Distributed ChecksumClearinghouse Software does not in any
+ * way grant permision to use Distributed Checksum and Reputation Clearinghouse
+ * software
+ *
+ * THE SOFTWARE IS PROVIDED "AS IS" AND RHYOLITE SOFTWARE, LLC DISCLAIMS ALL
  * WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES
- * OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL RHYOLITE SOFTWARE
+ * OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL RHYOLITE SOFTWARE, LLC
  * BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES
  * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
  * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
  * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
  * SOFTWARE.
  *
- * Rhyolite Software DCC 1.2.74-1.63 $Revision$
+ * Rhyolite Software DCC 1.3.42-1.80 $Revision$
  */
 
 #include "dccd_defs.h"
@@ -33,7 +50,7 @@ static RL rl_all_anon;			/* limit all or anonymous clients */
 static RL *rl_newest, *rl_oldest;
 static RL **rl_hash;
 static int rl_hash_len;
-static time_t rl_last_recycled;
+static u_char rl_too_many;
 
 time_t clients_cleared;
 
@@ -41,6 +58,7 @@ typedef struct ip_bl {
     struct ip_bl *fwd;
     struct in6_addr addr;
     struct in6_addr mask;
+    u_char  flags;
 } IP_BL;
 IP_BL *ip_bl;
 
@@ -48,20 +66,21 @@ IP_BL *ip_bl;
 
 /* See if this IP address should still be blacklisted
  * This should be sped up if there are ever more than a very few entries */
-static inline void
+static void
 ck_bl(RL *rl)
 {
 	const IP_BL *bl;
 
-	if (!(rl->flags & RL_CK_BLACKLIST)) {
-		rl->flags |= RL_CK_BLACKLIST;
-		for (bl = ip_bl; bl; bl = bl->fwd) {
-			if (DCC_IN_BLOCK(rl->clnt_addr, bl->addr, bl->mask)) {
-				rl->flags |= RL_BLACKLISTED;
-				return;
-			}
+	if (rl->flags & RL_CK_BLACKLIST)
+		return;
+	rl->flags |= RL_CK_BLACKLIST;
+	rl->flags &= ~(RL_BLACKLIST | RL_TRACE);
+
+	for (bl = ip_bl; bl; bl = bl->fwd) {
+		if (DCC_IN_BLOCK(rl->clnt_addr, bl->addr, bl->mask)) {
+			rl->flags |= bl->flags;
+			break;
 		}
-		rl->flags &= ~RL_BLACKLISTED;
 	}
 }
 
@@ -79,73 +98,79 @@ clear_bl(void)
 	}
 
 	for (rl = rl_newest; rl != 0; rl = rl->older) {
-		rl->flags  &= ~RL_CK_BLACKLIST;
+		rl->flags &= ~RL_CK_BLACKLIST;
 	}
 }
 
 
 
 void
-check_blacklist_file(u_char force)
+check_blacklist_file(void)
 {
 #define BL_NM "blacklist"
+	DCC_FNM_LNO_BUF fnm_buf;
 	static time_t prev_mtime;
-	static int serrno, syntax;
+	static int serrno;
 	struct stat sb;
 	FILE *f;
 	char buf[120];
 	IP_BL *bl;
 	struct in6_addr addr, mask;
+	u_char flags;
 	char *p;
-	int lineno, i;
+	int lno, entries, i;
+
+	/* not if a greylist server */
+	if (grey_on)
+		return;
 
 	/* see if the file has changed */
 	if (0 > stat(BL_NM, &sb)) {
 		if (errno != ENOENT) {
 			if (serrno != errno) {
 				serrno = errno;
-				syntax = 0;
 				dcc_error_msg("stat(%s): %s",
-					      DCC_NM2PATH(BL_NM), ERROR_STR());
+					      DB_NM2PATH_ERR(BL_NM),
+					      ERROR_STR());
 			}
-		} else if (ip_bl) {
-			prev_mtime = 0;
-			clear_bl();
+		} else if (prev_mtime != 0) {
+			dcc_trace_msg("%s disappeared",
+				      DB_NM2PATH_ERR(BL_NM));
 		}
+		clear_bl();
+		prev_mtime = 0;
 		return;
 	}
-	if (!force && prev_mtime == sb.st_mtime)
+	if (prev_mtime == sb.st_mtime)
 		return;
 
 	/* it has changed, so parse it */
+	clear_bl();
+	prev_mtime = 0;
 	f = fopen(BL_NM, "r");
 	if (!f) {
 		if (serrno != errno) {
 			serrno = errno;
-			syntax = 0;
 			dcc_error_msg("fopen(%s): %s",
-				      DCC_NM2PATH(BL_NM), ERROR_STR());
-		}
-		if (ip_bl) {
-			clear_bl();
+				      DB_NM2PATH_ERR(BL_NM), ERROR_STR());
 		}
+		return;
 	}
 	if (0 > fstat(fileno(f), &sb)) {
 		dcc_error_msg("fstat(%s): %s",
-			      DCC_NM2PATH(BL_NM), ERROR_STR());
+			      DB_NM2PATH_ERR(BL_NM), ERROR_STR());
 		return;
 	}
 	prev_mtime = sb.st_mtime;
 
-	clear_bl();
-	for (lineno = 1; ; ++lineno) {
+	entries = 0;
+	for (lno = 1; ; ++lno) {
 		if (!fgets(buf, sizeof(buf), f)) {
 			if (ferror(f)
 			    && serrno != errno) {
 				serrno = errno;
-				syntax = 0;
 				dcc_error_msg("fgets(%s): %s",
-					      DCC_NM2PATH(BL_NM),
+					      DB_NM2PATH_ERR(BL_NM),
 					      ERROR_STR());
 			}
 			break;
@@ -153,12 +178,9 @@ check_blacklist_file(u_char force)
 		/* reject lines that are too long */
 		i = strlen(buf);
 		if (buf[i-1] != '\n') {
-			if (syntax != lineno) {
-				serrno = 0;
-				syntax = lineno;
-				dcc_error_msg("syntax error%s",
-					      fnm_lineno(BL_NM, lineno));
-			}
+			dcc_error_msg("syntax error%s",
+				      fnm_lno(fnm_buf,
+					      DB_NM2PATH_ERR(BL_NM), lno));
 			break;
 		}
 
@@ -166,32 +188,53 @@ check_blacklist_file(u_char force)
 		p = strchr(buf, '#');
 		if (p)
 			*p = '\0';
-		i = strspn(buf, DCC_WHITESPACE);
-		if (buf[i] == '\0')
+		p = buf;
+		p += strspn(p, DCC_WHITESPACE);
+		if (*p == '\0')
 			continue;
 
-		i = dcc_str2cidr(dcc_emsg, &addr, &mask, &buf[i],
-				 BL_NM, lineno);
-		if (i <= 0) {
-			if (syntax != lineno) {
-				serrno = 0;
-				syntax = lineno;
-				if (i < 0)
-					dcc_error_msg("%s", dcc_emsg);
-				else
-					dcc_error_msg("syntax error%s",
-						      fnm_lineno(BL_NM,lineno));
+		flags = RL_BLACKLIST;
+		for (;;) {
+			if (!CSTRCMP(p, "trace")
+			    && 0 != (i = strspn(p+STRZ("trace"), " \t"))) {
+				p += i+STRZ("trace");
+				flags |= RL_TRACE;
+				continue;
 			}
+			if (!CSTRCMP(p, "ok")
+			    && 0 != (i = strspn(p+STRZ("ok"), " \t"))) {
+				p += i+STRZ("ok");
+				flags &= ~RL_BLACKLIST;
+				continue;
+			}
+			break;
+		}
+
+		i = dcc_str2cidr(dcc_emsg, &addr, &mask, p, BL_NM, lno);
+		if (i <= 0) {
+			if (i < 0)
+				dcc_error_msg("%s", dcc_emsg);
+			else
+				dcc_error_msg("syntax error%s",
+					      fnm_lno(fnm_buf,
+						      DB_NM2PATH_ERR(BL_NM),
+						      lno));
 			break;
 		}
 
 		bl = dcc_malloc(sizeof(*bl));
 		bl->addr = addr;
 		bl->mask = mask;
+		bl->flags = flags;
 		bl->fwd = ip_bl;
 		ip_bl = bl;
+		++entries;
 	}
 	fclose(f);
+
+	if (entries)
+		dcc_trace_msg("read %d entries from %s",
+			      entries, DB_NM2PATH_ERR(BL_NM));
 }
 
 
@@ -280,28 +323,29 @@ rl_age(RL *rl, const RL_RATE *credits)
 	secs = (db_time.tv_sec - rl->last_used);
 
 	/* only prevent overflow if no time has passed */
-	if (secs != 0) {
-		rl->last_used = db_time.tv_sec;
-
-		/* reset idle counters */
-		if (secs >= RL_AVG_SECS || secs < 0) {
-			rl->bug_credits = rl_bugs_rate.hi;
-			rl->request_credits = credits->hi;
-			return;
-		}
-
-		rl->request_credits += secs*credits->sec;
-		if (rl->request_credits > credits->hi)
-			rl->request_credits = credits->hi;
-
-		rl->bug_credits += secs*rl_bugs_rate.sec;
-		if (rl->bug_credits > rl_bugs_rate.hi)
-			rl->bug_credits = rl_bugs_rate.hi;
+	if (secs == 0) {
+		if (rl->request_credits < credits->lo)
+			rl->request_credits = credits->lo;
+		if (rl->bug_credits < rl_bugs_rate.lo)
+			rl->bug_credits = rl_bugs_rate.lo;
 	}
-	if (rl->request_credits < credits->lo)
-		rl->request_credits = credits->lo;
-	if (rl->bug_credits < rl_bugs_rate.lo)
-		rl->bug_credits = rl_bugs_rate.lo;
+
+	rl->last_used = db_time.tv_sec;
+
+	/* reset idle counters */
+	if (secs >= RL_AVG_SECS || secs < 0) {
+		rl->bug_credits = rl_bugs_rate.hi;
+		rl->request_credits = credits->hi;
+		return;
+	}
+
+	rl->request_credits += secs*credits->sec;
+	if (rl->request_credits > credits->hi)
+		rl->request_credits = credits->hi;
+
+	rl->bug_credits += secs*rl_bugs_rate.sec;
+	if (rl->bug_credits > rl_bugs_rate.hi)
+		rl->bug_credits = rl_bugs_rate.hi;
 }
 
 
@@ -346,44 +390,63 @@ rl_ref(RL *rl)
 static RL *
 rl_get_free(void)
 {
-	RL *rl;
+	RL *rl, *rl2;
 	time_t stale;
 
 	for (rl = rl_oldest; rl != 0; rl = rl->newer) {
-		if (!rl->ref_cnt) {
-			/* Found oldest free block */
-			if (rl->last_used != 0) {
-				/* it has been used */
-				stale = db_time.tv_sec;
-				/* keep a day's worth of blocks until
-				 * we get enough to worry about a denial
-				 * of service attack */
-				if (rl_hash_len < RL_MIN_MAX)
-					stale -= CLIENTS_AGE;
-				else
-					stale -= RL_LIFE_SECS;
-				/* make more if the oldest is new */
-				if (rl->last_used >= stale
-				    && rl_hash_len <= RL_MAX_MAX)
-					return 0;
-				/* Remember how old it was */
-				rl_last_recycled = rl->last_used;
+		if (rl->ref_cnt)
+			continue;
+
+		/* Found oldest free block */
+		if (rl->last_used != 0) {
+			/* it has been used */
+			stale = db_time.tv_sec;
+			/* keep a day's worth of blocks until
+			 * we get enough to worry about a denial
+			 * of service attack */
+			if (rl_hash_len < RL_MIN_MAX)
+				stale -= CLIENTS_AGE;
+			else
+				stale -= RL_LIFE_SECS;
+			/* make more if the oldest is new */
+			if (rl->last_used >= stale
+			    && rl_hash_len <= RL_MAX_MAX)
+				return 0;
+
+			/* Notice if we are about to recycle a block that is
+			 * not obsolete.
+			 * Try to find an old, little used block, so that we
+			 * do not recycle a block used by a busy client that
+			 * has just paused */
+			if (rl->last_used > clients_cleared) {
+				rl_too_many = 1;
+				stale = db_time.tv_sec - (CLIENTS_AGE*3)/4;
+				for (rl2 = rl;
+				     rl2 && rl2->last_used <= stale;
+				     rl2 = rl2->newer) {
+					if (rl2->requests < 100
+					    && RL_REQUESTS_AVG(rl) < 100
+					    && !(rl->flags & RL_CK_BLACKLIST)) {
+					    rl = rl2;
+					    break;
+					}
+				}
 			}
-
-			/* recycle the oldest free block */
-			if (rl->fwd)
-				rl->fwd->bak = rl->bak;
-			if (rl->bak)
-				rl->bak->fwd = rl->fwd;
-			else if (rl->hash)
-				*rl->hash = rl->fwd;
-			rl_unref(rl);
-			memset(rl, 0, sizeof(*rl));
-
-			rl->requests_avg_date = db_time.tv_sec + RL_AVG_DAY;
-			rl_ref(rl);
-			return rl;
 		}
+
+		/* recycle a block */
+		if (rl->fwd)
+			rl->fwd->bak = rl->bak;
+		if (rl->bak)
+			rl->bak->fwd = rl->fwd;
+		else if (rl->hash)
+			*rl->hash = rl->fwd;
+		rl_unref(rl);
+		memset(rl, 0, sizeof(*rl));
+
+		rl->requests_avg_date = db_time.tv_sec + RL_AVG_DAY;
+		rl_ref(rl);
+		return rl;
 	}
 
 	/* there are no free blocks that are old enough to recycle */
@@ -468,109 +531,116 @@ rl_get(QUEUE *q)
 
 
 
-/* get the most recent clients */
-int					/* +/- number of clients */
-clients_get(DCC_ADMN_RESP_VAL *val,
-	    int *lenp,			/* buffer length */
-	    u_int offset,		/* skip this many newer entries */
-	    int thold,			/* skip clients with fewer requests */
-	    u_char avg)			/* 1=get average requests */
+static u_char *
+client_pack4(u_char *cp,
+	     u_int32_t v)
 {
+	while (v > 0x7f) {
+		*cp++ = v | 0x80;
+		v >>= 7;
+	}
+	*cp++ = v;
+	return cp;
+}
+
+
+
+static int
+client_pack(u_char *cp0,
+	    u_char **pflags,
+#ifdef DCC_PKT_VERSION6
+	     const QUEUE *q,
+#endif
+	     u_char flags,
+	     DCC_CLNT_ID clnt_id,
+	     time_t last_used,		/* skip count if *_CLIENTS_SKIP */
+	     int requests,
+	     int nops,
+	     u_char vers,
+	     const struct in6_addr *clnt_addr)
+{
+	u_char *cp;
+
+#ifdef DCC_PKT_VERSION6
+	if (q && q->pkt.hdr.pkt_vers <= DCC_PKT_VERSION6) {
 #define CPY2(t,s) (t[0] = s>>8, t[1] = s)
 #define CPY3(t,s) (t[0] = s>>16, t[1] = s>>8, t[2] = s)
 #define CPY4(t,s) (t[0] = s>>24, t[1] = s>>16, t[2] = s>>8, t[3] = s)
-	RL *rl;
-	DCC_ADMN_RESP_CLIENTS *cl, *cl_skip;
-	int requests, nops;
-	int skipped, total, len, len_lim;
+		DCC_ADMN_RESP_CLIENTSv6 *cl;
+		int addr_len;
 
-	if (!val || !lenp) {
-		len_lim = 0;
-	} else {
-		len_lim = *lenp;
-	}
-	if (!thold)
-		thold = 1;
-	cl_skip = 0;
-	skipped = 0;
-	total = 0;
-	len = 0;
-	for (rl = rl_newest; rl != 0; rl = rl->older) {
-		if (rl->last_used == 0)
-			continue;
-
-		if (rl->requests != 0)
-			++total;
-		/* get the part of the list that cdcc wants */
-		if (offset != 0) {
-			--offset;
-			continue;
-		}
-
-		if (len+ISZ(*cl) > len_lim) {
-			if (len_lim != 0)
-				break;
-			continue;
-		}
-
-		if (avg) {
-			rl_avg_requests_age(rl, 0);
-			requests = RL_REQUESTS_AVG(rl);
-			nops = RL_NOPS_AVG(rl);
+		cl = (DCC_ADMN_RESP_CLIENTSv6 *)cp0;
+		*pflags = 0;
+		cl->flags = flags & (DCC_ADMN_RESP_CLIENTS_BL
+				     | DCC_ADMN_RESP_CLIENTS_IPV6
+				     | DCC_ADMN_RESP_CLIENTS_SKIP);
+		if (flags & DCC_ADMN_RESP_CLIENTS_SKIP) {
+			CPY4(cl->last_used, 0);
+			CPY3(cl->requests, last_used);
 		} else {
-			requests = rl->requests;
-			nops = rl->nops;
+			CPY4(cl->last_used, last_used);
+			CPY3(cl->requests, requests);
 		}
-		if (requests < thold
-		    && !(rl->flags & RL_BLACKLISTED)) {
-			/* insert a fake entry for skipped records */
-			if (!cl_skip) {
-				cl_skip = (DCC_ADMN_RESP_CLIENTS *
-					   )((char *)val->clients + len);
-				memset(cl_skip, 0, sizeof(*cl_skip));
-				cl_skip->flags = DCC_ADMN_RESP_CLIENTS_SKIP;
-				len += (sizeof(*cl_skip)
-					- sizeof(cl_skip->addr)
-					+ sizeof(cl_skip->addr.ipv4));
-			}
-			++skipped;
-			CPY3(cl_skip->requests, skipped);
-			continue;
-		}
-
-		cl = (DCC_ADMN_RESP_CLIENTS *)((char *)val->clients + len);
-		cl->flags = 0;
-		CPY4(cl->last_used, rl->last_used);
-		CPY3(cl->requests, requests);
 		CPY2(cl->nops, nops);
-		CPY4(cl->id, rl->clnt_id);
-		if (rl->flags & RL_BLACKLISTED)
-			cl->flags |= DCC_ADMN_RESP_CLIENTS_BL;
-		if (DCC_IN6_ADDR_V4MAPPED(&rl->clnt_addr)) {
-			memcpy(&cl->addr.ipv4,
-			       &rl->clnt_addr.s6_addr32[3],
-			       sizeof(cl->addr.ipv4));
-			len += (sizeof(*cl) - sizeof(cl->addr)
-				+ sizeof(cl->addr.ipv4));
+		CPY4(cl->clnt_id, clnt_id);
+		if (!clnt_addr) {
+			addr_len = 4;
+			memset(&cl->addr, 0, addr_len);
+		} else if (DCC_IN6_ADDR_V4MAPPED(clnt_addr)) {
+			addr_len = 4;
+			memcpy(&cl->addr, &clnt_addr->s6_addr32[3], addr_len);
 		} else {
 			cl->flags |= DCC_ADMN_RESP_CLIENTS_IPV6;
-			memcpy(&cl->addr.ipv6,
-			       rl->clnt_addr.s6_addr32,
-			       sizeof(cl->addr.ipv6));
-			len += (sizeof(*cl) - sizeof(cl->addr)
-				+ sizeof(cl->addr.ipv6));
+			addr_len = 16;
+			memcpy(&cl->addr, clnt_addr, addr_len);
 		}
-	}
-	if (lenp)
-		*lenp = len;
 
-	/* return negative total if we have more clients than RL blocks */
-	if (rl_last_recycled > clients_cleared)
-		return -total;
-	return total;
+		return sizeof(*cl) - sizeof(cl->addr) + addr_len;
 #undef CPY2
 #undef CPY3
 #undef CPY4
+	}
+#endif
+
+	cp = cp0;
+	*pflags = cp0;
+	if (vers != 0)
+		flags |= DCC_ADMN_RESP_CLIENTS_VERS;
+	*cp++ = flags;
+	if (flags & DCC_ADMN_RESP_CLIENTS_VERS)
+		*cp++ = vers;
+	*cp++ = last_used >> 24;
+	*cp++ = last_used >> 16;
+	*cp++ = last_used >> 8;
+	*cp++ = last_used;
+	if (clnt_id == DCC_ID_ANON) {
+		*cp0 |= DCC_ADMN_RESP_CLIENTS_ID1;
+	} else {
+		cp = client_pack4(cp, clnt_id);
+	}
+	cp = client_pack4(cp, requests);
+	cp = client_pack4(cp, nops);
+	if (!clnt_addr) {
+		memset(cp, 0, 4);
+		cp += 4;
+	} else if (DCC_IN6_ADDR_V4MAPPED(clnt_addr)) {
+		memcpy(cp, &clnt_addr->s6_addr32[3], 4);
+		cp += 4;
+	} else {
+		*cp0 |= DCC_ADMN_RESP_CLIENTS_IPV6;
+		memcpy(cp, clnt_addr->s6_addr32, 16);
+		cp += 16;
+	}
+	return cp-cp0;
+}
+
+
+
+static int
+client_pack_skip(u_char *cp, u_char **pflags, const QUEUE *q, u_int skipped)
+{
+	return client_pack(cp, pflags, q, DCC_ADMN_RESP_CLIENTS_SKIP,
+			   DCC_ID_ANON, skipped, 0, 0, 0, 0);
 }
 
 
@@ -578,88 +648,234 @@ clients_get(DCC_ADMN_RESP_VAL *val,
 void
 clients_get_id(DCC_ADMN_RESP_VAL *val,
 	       int *lenp,		/* buffer length */
+	       const QUEUE *q,
 	       u_int offset,		/* skip this many newer entries */
 	       int thold,		/* skip clients with fewer requests */
-	       u_char avg)		/* 1=get average requests */
+	       u_char req_flags,
+	       const struct in6_addr *addr6,
+	       const struct in6_addr *mask6)
 {
-#define CPY2(t,s) (t[0] = s>>8, t[1] = s)
-#define CPY3(t,s) (t[0] = s>>16, t[1] = s>>8, t[2] = s)
-#define CPY4(t,s) (t[0] = s>>24, t[1] = s>>16, t[2] = s>>8, t[3] = s)
 	RL *rl, *rl2;
-	DCC_ADMN_RESP_CLIENTS *cl, *cl_skip;
+	u_char *skip;
 	int requests, nops;
+	u_char *pflags;
 	u_int skipped;
 	int len, len_lim;
 
+	pflags = 0;
 	len_lim = *lenp;
-	cl_skip = 0;
+#ifdef DCC_PKT_VERSION6
+	if (q->pkt.hdr.pkt_vers <= DCC_PKT_VERSION6)
+		len_lim *= sizeof(DCC_ADMN_RESP_CLIENTSv6);
+	else
+#endif
+		len_lim <<= DCC_ADMIN_RESP_CLIENTS_SHIFT;
+	if (len_lim > ISZ(*val))
+		len_lim = ISZ(*val);
+	skip = 0;
 	skipped = 0;
 	len = 0;
 	if (!thold)
 		thold = 1;
-	for (rl = rl_newest; rl != 0; rl = rl->older) {
+	for (rl = rl_newest; rl != 0; rl = rl->older)
 		rl->flags &= ~RL_MARKED;
-	}
-	for (rl = rl_newest; rl != 0; rl = rl->older) {
+
+	for (rl = rl_newest; ; rl = rl->older) {
+		if (!rl) {
+			if (pflags)
+				*pflags |= DCC_ADMN_RESP_CLIENTS_LAST;
+			break;
+		}
+
 		if (rl->last_used == 0)
 			continue;
 
+		if (rl->flags & RL_MARKED)
+			continue;
+
+		if (addr6 && !DCC_IN_BLOCK(rl->clnt_addr, *addr6, *mask6)) {
+			rl->flags |= RL_MARKED;
+			continue;
+		}
+
+		requests = 0;
+		nops = 0;
+		for (rl2 = rl; rl2 != 0; rl2 = rl2->older) {
+			if (rl2->clnt_id != rl->clnt_id)
+				continue;
+			rl2->flags |= RL_MARKED;
+			if (addr6
+			    && !DCC_IN_BLOCK(rl->clnt_addr, *addr6, *mask6))
+				continue;
+			if (req_flags & DCC_AOP_CLIENTS_AVG) {
+				rl_avg_requests_age(rl2, 0);
+				requests += RL_REQUESTS_AVG(rl2);
+				nops += RL_NOPS_AVG(rl2);
+			} else {
+				requests += rl2->requests;
+				nops += rl2->nops;
+			}
+		}
+
 		/* get the part of the list that cdcc wants */
 		if (offset != 0) {
 			--offset;
 			continue;
 		}
 
-		requests = 0;
-		nops = 0;
-		if (!(rl->flags & RL_MARKED)) {
-			for (rl2 = rl; rl2 != 0; rl2 = rl2->older) {
-				if (rl2->clnt_id == rl->clnt_id) {
-					if (avg) {
-					    rl_avg_requests_age(rl2, 0);
-					    requests += RL_REQUESTS_AVG(rl2);
-					    nops += RL_NOPS_AVG(rl2);
-					} else {
-					    requests += rl2->requests;
-					    nops += rl2->nops;
-					}
-					rl2->flags |= RL_MARKED;
-				}
-			}
-		}
 		if (requests < thold) {
-			/* insert a fake entry for skipped records */
-			if (!cl_skip) {
-				cl_skip = (DCC_ADMN_RESP_CLIENTS *
-					   )((char *)val->clients + len);
-				memset(cl_skip, 0, sizeof(*cl_skip));
-				cl_skip->flags = DCC_ADMN_RESP_CLIENTS_SKIP;
-				len += (sizeof(*cl_skip)
-					- sizeof(cl_skip->addr)
-					+ sizeof(cl_skip->addr.ipv4));
+			/* The threshold might be larger on the next request
+			 * from cdcc, so tell cdcc the number skipped for
+			 * the threshold this time.
+			 * Tell cdcc by insertint a fake entry. */
+			if (!skip) {
+				skip = &val->clients[len];
+				len += client_pack_skip(skip, &pflags, q, 0);
 			}
 			++skipped;
-			CPY3(cl_skip->requests, skipped);
 			continue;
 		}
 
-		cl = (DCC_ADMN_RESP_CLIENTS *)((char *)val->clients + len);
-		cl->flags = 0;
-		CPY4(cl->last_used, rl->last_used);
-		CPY3(cl->requests, requests);
-		CPY2(cl->nops, nops);
-		CPY4(cl->id, rl->clnt_id);
-		memset(&cl->addr.ipv4, 0, sizeof(cl->addr.ipv4));
-		len += (sizeof(*cl) - sizeof(cl->addr)
-			+ sizeof(cl->addr.ipv4));
-
-		if (len+ISZ(*cl) > len_lim)
+		/* stop at end of buffer
+		 * check only after skipping boring records for the common
+		 * case of the last 10,000 records missing the threshold */
+		if (len+DCC_ADMN_RESP_CLIENTS_MAX_SIZE > len_lim)
 			break;
+
+		len += client_pack(&val->clients[len], &pflags, q, 0,
+				   rl->clnt_id, rl->last_used,
+				   requests, nops, 0, 0);
 	}
+	if (skipped)
+		client_pack_skip(skip, &pflags, q, skipped);
 	*lenp = len;
-#undef CPY2
-#undef CPY3
-#undef CPY4
+}
+
+
+
+/* get the most recent clients */
+int					/* +/- number of clients */
+clients_get(DCC_ADMN_RESP_VAL *val,
+	    int *lenp,			/* buffer length */
+#ifdef DCC_PKT_VERSION6
+	    const QUEUE *q,
+#endif
+	    u_int offset,		/* skip this many newer entries */
+	    int thold,			/* skip clients with fewer requests */
+	    u_char req_flags,
+	    const struct in6_addr *addr6,
+	    const struct in6_addr *mask6)
+{
+	RL *rl;
+	u_char *skip;
+	int requests, nops;
+	u_char prev_vers, vers, *pflags;
+	int skipped, total, len, len_lim;
+
+	pflags = 0;
+	if (!val || !lenp) {
+		len_lim = 0;
+	} else {
+		len_lim = *lenp;
+#ifdef DCC_PKT_VERSION6
+		if (q && q->pkt.hdr.pkt_vers <= DCC_PKT_VERSION6)
+			len_lim *= sizeof(DCC_ADMN_RESP_CLIENTSv6);
+		else
+#endif
+		len_lim <<= DCC_ADMIN_RESP_CLIENTS_SHIFT;
+		if (len_lim > ISZ(*val))
+			len_lim = ISZ(*val);
+	}
+	if (!thold)
+		thold = 1;
+	prev_vers = 0;
+	skip = 0;
+	skipped = 0;
+	total = 0;
+	len = 0;
+	for (rl = rl_newest; ; rl = rl->older) {
+		if (!rl) {
+			if (pflags)
+				*pflags |= DCC_ADMN_RESP_CLIENTS_LAST;
+			break;
+		}
+
+		if (rl->last_used == 0)
+			continue;
+
+		if (rl->requests != 0)
+			++total;
+
+		/* compute only the total for `cdcc stats` if buffer is null */
+		if (len_lim == 0)
+			continue;
+
+		if (addr6 && !DCC_IN_BLOCK(rl->clnt_addr, *addr6, *mask6))
+			continue;
+
+		if (offset != 0) {
+			--offset;
+			continue;
+		}
+
+		if (req_flags & DCC_AOP_CLIENTS_AVG) {
+			rl_avg_requests_age(rl, 0);
+			requests = RL_REQUESTS_AVG(rl);
+			nops = RL_NOPS_AVG(rl);
+		} else {
+			requests = rl->requests;
+			nops = rl->nops;
+		}
+
+		/* skip uninteresting records and insert a fake
+		 * entry in the output to the client with the total
+		 * skipped in the entire response to the client */
+		if (requests < thold
+		    && (!(rl->flags & RL_BLACKLIST)
+			|| rl->requests == 0)) {
+			if (!skip) {
+				/* start a new fake record */
+				skip = &val->clients[len];
+				len += client_pack_skip(skip, &pflags, q, 0);
+				prev_vers = 0;
+			}
+			++skipped;
+			continue;
+		}
+
+		/* stop at end of buffer
+		 * check only after skipping boring records for the common
+		 * case of ignoring the last 10,000 records */
+		if (len + DCC_ADMN_RESP_CLIENTS_MAX_SIZE*2 > len_lim)
+				break;
+
+		/* send the version number if it is wanted and differs
+		 * from the previous value */
+		if ((req_flags & DCC_AOP_CLIENTS_VERS)
+		    && rl->vers != prev_vers) {
+			vers = rl->vers;
+			prev_vers = vers;
+		} else {
+			vers = 0;
+		}
+
+		len += client_pack(&val->clients[len], &pflags, q,
+				   (rl->flags & RL_BLACKLIST)
+				   ? DCC_ADMN_RESP_CLIENTS_BL : 0,
+				   rl->clnt_id, rl->last_used,
+				   requests, nops, vers, &rl->clnt_addr);
+	}
+	/* put final total number of skipped records in the output */
+	if (skipped)
+		client_pack_skip(skip, &pflags, q, skipped);
+
+	if (lenp)
+		*lenp = len;
+
+	/* return negative total if we have more clients than RL blocks */
+	if (rl_too_many)
+		return -total;
+	return total;
 }
 
 
@@ -679,6 +895,7 @@ clients_clear(u_char force_rl_age)
 	}
 
 	clients_cleared = db_time.tv_sec;
+	rl_too_many = 0;
 }
 
 
@@ -775,13 +992,23 @@ ck_id(QUEUE *q, DCC_CLNT_ID id)
 			q->flags |= Q_FLG_RPT_OK;
 
 		rl = rl_get(q);
+		rl->vers = q->pkt.hdr.pkt_vers;
 		++rl->requests;
 		rl_age(rl, &rl_sub_rate);
 		rl->request_credits -= RL_SCALE;
 
-		if (rl->flags & RL_BLACKLISTED) {
-			TMSG2(BL, "%d %s blacklisted request",
-			      id, Q_CIP(q));
+		if (rl->flags & RL_TRACE)
+			dcc_trace_msg("%d %s%s %s",
+				      id, Q_CIP(q),
+				      (rl->flags & RL_BLACKLIST)
+				      ? " blacklisted" : "",
+				      dcc_hdr_op2str(0, 0,
+						     &q->pkt.hdr));
+		if (rl->flags & RL_BLACKLIST) {
+			if (!(rl->flags & RL_TRACE))
+				TMSG3(BL, "%d %s blacklisted %s",
+				      id, Q_CIP(q), dcc_hdr_op2str(0, 0,
+							&q->pkt.hdr));
 			return 0;
 		}
 
@@ -803,13 +1030,23 @@ ck_id(QUEUE *q, DCC_CLNT_ID id)
 		rl_age(&rl_all_anon, &rl_all_anon_rate);
 		rl_all_anon.request_credits -= RL_SCALE;
 		rl = rl_get(q);
+		rl->vers = q->pkt.hdr.pkt_vers;
 		++rl->requests;
 		rl_age(rl, &rl_anon_rate);
 		rl->request_credits -= RL_SCALE;
 
-		if (rl->flags & RL_BLACKLISTED) {
-			TMSG1(BL, " anonymous %s blacklisted request",
-			      Q_CIP(q));
+		if (rl->flags & RL_TRACE)
+			dcc_trace_msg("anonymous %s%s %s",
+				      Q_CIP(q),
+				      (rl->flags & RL_BLACKLIST)
+				      ? " blacklisted" : "",
+				      dcc_hdr_op2str(0, 0,
+						     &q->pkt.hdr));
+		if (rl->flags & RL_BLACKLIST) {
+			if (!(rl->flags & RL_TRACE))
+				TMSG2(BL, "anonymous %s blacklisted %s",
+				      Q_CIP(q), dcc_hdr_op2str(0, 0,
+							&q->pkt.hdr));
 			return 0;
 		}
 
@@ -846,9 +1083,10 @@ ck_clnt_srvr_id(QUEUE *q)
 {
 	DCC_CLNT_ID id;
 	u_char result;
+	char ob[DCC_OPBUF];
 
 	/* require a client-ID, our server-ID, or the anonymous client-ID
-	 * to allow anonymous `cdcc stats` */
+	 * to consider allowing an administrative request */
 	id = ntohl(q->pkt.hdr.sender);
 	if ((id < DCC_SRVR_ID_MIN || id > DCC_CLNT_ID_MAX)
 	    && id != DCC_ID_ANON) {
@@ -856,7 +1094,8 @@ ck_clnt_srvr_id(QUEUE *q)
 		result = ck_id(q, DCC_ID_ANON);
 		if (result)
 			anon_msg("bad client or server-ID %d from %s for %s",
-				 id, Q_CIP(q), dcc_req_op2str(&q->pkt.a));
+				 id, Q_CIP(q),
+				 dcc_hdr_op2str(ob, sizeof(ob), &q->pkt.hdr));
 		return result;
 	}
 	return ck_id(q, id);
@@ -871,6 +1110,7 @@ ck_clnt_id(QUEUE *q)
 {
 	DCC_CLNT_ID id;
 	u_char result;
+	char ob[DCC_OPBUF];
 
 	/* require a client-ID instead of a server-ID to discourage server
 	 * operators from leaking server-ID's */
@@ -881,7 +1121,8 @@ ck_clnt_id(QUEUE *q)
 		result = ck_id(q, DCC_ID_ANON);
 		if (result)
 			anon_msg("bad client-ID %d from %s for %s",
-				 id, Q_CIP(q), dcc_req_op2str(&q->pkt.a));
+				 id, Q_CIP(q),
+				 dcc_hdr_op2str(ob, sizeof(ob), &q->pkt.hdr));
 		return result;
 	}
 	return ck_id(q, id);
