@@ -2,25 +2,42 @@
  *
  * decode MIME for checksums
  *
- * Copyright (c) 2005 by Rhyolite Software
+ * Copyright (c) 2005 by Rhyolite Software, LLC
+ *
+ * This agreement is not applicable to any entity which sells anti-spam
+ * solutions to others or provides an anti-spam solution as part of a
+ * security solution sold to other entities, or to a private network
+ * which employs the DCC or uses data provided by operation of the DCC
+ * but does not provide corresponding data to other users.
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
  * copyright notice and this permission notice appear in all copies.
  *
- * THE SOFTWARE IS PROVIDED "AS IS" AND RHYOLITE SOFTWARE DISCLAIMS ALL
+ * Parties not eligible to receive a license under this agreement can
+ * obtain a commercial license to use DCC and permission to use
+ * U.S. Patent 6,330,590 by contacting Commtouch at http://www.commtouch.com/
+ * or by email to nospam@commtouch.com.
+ *
+ * A commercial license would be for Distributed Checksum and Reputation
+ * Clearinghouse software.  That software includes additional features.  This
+ * free license for Distributed ChecksumClearinghouse Software does not in any
+ * way grant permision to use Distributed Checksum and Reputation Clearinghouse
+ * software
+ *
+ * THE SOFTWARE IS PROVIDED "AS IS" AND RHYOLITE SOFTWARE, LLC DISCLAIMS ALL
  * WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES
- * OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL RHYOLITE SOFTWARE
+ * OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL RHYOLITE SOFTWARE, LLC
  * BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES
  * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
  * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
  * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
  * SOFTWARE.
  *
- * Rhyolite Software DCC 1.2.74-1.30 $Revision$
+ * Rhyolite Software DCC 1.3.42-1.37 $Revision$
  */
 
-#include "ck.h"
+#include "dcc_ck.h"
 
 /* Notice MIME headers */
 void
@@ -585,11 +602,11 @@ dcc_ck_url(DCC_URL_SKIP *url, char c, char **pbufp)
 	switch (url->st) {
 	case DCC_URL_ST_IDLE:
 		if (c == 'h') {
-			/* look for 't' after 'h' in "http" */
+			/* start looking for 't' after 'h' in "http" */
 			url->flags = 0;
 			url->st = DCC_URL_ST_T1;
 		} else if (c == '=') {
-			/* look for the '=' in "href=" or "img=" */
+			/* look for the '=' in "href=" or "img src=" */
 			url->st = DCC_URL_ST_QUOTE;
 		}
 		RET_C(DCC_CK_URL_CHAR);
@@ -597,9 +614,13 @@ dcc_ck_url(DCC_URL_SKIP *url, char c, char **pbufp)
 	case DCC_URL_ST_QUOTE:
 		/* look for '"' or 'H' after "href=" */
 		if (c == 'h') {
-			url->flags = 0;
+			url->flags &= ~DCC_URL_QUOTES;
 			url->st = DCC_URL_ST_T1;
 		} else if (c == '"') {
+			url->flags |= DCC_URL_DQUOTED;
+			url->st = DCC_URL_ST_QH;
+		} else if (c == '\'') {
+			url->flags |= DCC_URL_SQUOTED;
 			url->st = DCC_URL_ST_QH;
 		} else {
 			url->st = DCC_URL_ST_IDLE;
@@ -607,8 +628,8 @@ dcc_ck_url(DCC_URL_SKIP *url, char c, char **pbufp)
 		RET_C(DCC_CK_URL_CHAR);
 
 	case DCC_URL_ST_QH:
+		/* seen quote; looking for start of URL */
 		if (c == 'h') {
-			url->flags = DCC_URL_QUOTED;
 			url->st = DCC_URL_ST_T1;
 		} else {
 			url->st = DCC_URL_ST_IDLE;
@@ -675,6 +696,8 @@ dcc_ck_url(DCC_URL_SKIP *url, char c, char **pbufp)
 		url->dot = 0;
 		url->start = *pbufp;
 		url->total = 0;
+		url->flags &= ~(DCC_URL_DEL_DOMAIN
+				| DCC_URL_PERCENT1 | DCC_URL_PERCENT2);
 		url->st = DCC_URL_ST_SLASH3;
 		/* fall into DCC_URL_ST_SLASH3 */
 	case DCC_URL_ST_SLASH3:
@@ -683,9 +706,15 @@ dcc_ck_url(DCC_URL_SKIP *url, char c, char **pbufp)
 again:
 		if (c == '.') {
 			/* keep only 1st and 2nd level domain names */
-			url->flags &= ~DCC_URL_SIMPLE;
+			url->flags &= ~DCC_URL_DEL_DOMAIN;
 			if (!url->dot) {
-				url->dot = *pbufp;
+				/* do nothing at first '.' unless the name
+				 * was too long */
+				if (*pbufp >= url->start+DCC_URL_MAX) {
+					*pbufp = url->start;
+				} else {
+					url->dot = *pbufp;
+				}
 			} else {
 				url->flags |= DCC_URL_DEL_DOMAIN;
 			}
@@ -705,27 +734,34 @@ again:
 			url->st = DCC_URL_ST_SKIP;
 			RET_C(DCC_CK_URL_HOST_END);
 		}
-		if (c == '"') {
+		if (c == '"' && (url->flags & DCC_URL_DQUOTED)) {
 			url->st = DCC_URL_ST_IDLE;
 			RET_C(DCC_CK_URL_HOST_END);
 		}
-		if (c == '>' || c == ' ' || c == '\t'
-		    || c == '\n' || c == '\r') {
+		if (c == '\'' && (url->flags & DCC_URL_SQUOTED)) {
+			url->st = DCC_URL_ST_IDLE;
+			RET_C(DCC_CK_URL_HOST_END);
+		}
+		if ((c == '<' || c == '>')
+		    && (url->flags & DCC_URL_QUOTES) == 0) {
+			url->st = DCC_URL_ST_IDLE;
+			RET_C(DCC_CK_URL_HOST_END);
+		}
+		if (c == ' ' || c == '\t' || c == '\n' || c == '\r') {
 			if (!(url->flags & DCC_URL_QUOTED)
 			    || url->total > DCC_URL_FAILSAFE) {
 				url->st = DCC_URL_ST_IDLE;
 				RET_C(DCC_CK_URL_HOST_END);
 			}
 			/* whitespace in a URL hostname is at best username */
-			url->dot = 0;
-			url->flags &= ~DCC_URL_SIMPLE;
+			*pbufp = url->start;
+			url->st = DCC_URL_ST_SLASH3_START;
 			RET_C(DCC_CK_URL_HOST_RESET);
 		}
 		if (c == '@') {
 			/* ignore username and password */
 			*pbufp = url->start;
-			url->dot = 0;
-			url->flags &= ~DCC_URL_SIMPLE;
+			url->st = DCC_URL_ST_SLASH3_START;
 			RET_C(DCC_CK_URL_HOST_RESET);
 		}
 
@@ -740,11 +776,13 @@ again:
 			} else if (c >= 'a' && c <= 'f') {
 				c -= 'a'-10;
 			} else {
-				url->flags &= ~DCC_URL_SIMPLE;
+				*pbufp = url->start;
+				url->st = DCC_URL_ST_SLASH3_START;
 				RET_C(DCC_CK_URL_HOST_RESET);
 			}
 			if (url->flags & DCC_URL_PERCENT2) {
-				url->flags &= ~DCC_URL_SIMPLE;
+				url->flags &= ~(DCC_URL_PERCENT1
+						| DCC_URL_PERCENT2);
 				c |= url->percent;
 				if (c >= 'A' && c <= 'Z')
 					c -= 'A' - 'a';
@@ -768,13 +806,13 @@ again:
 	case DCC_URL_ST_SKIP:
 		/* skip the rest of the URL */
 		++url->total;
-		if (c == '"') {
+		if (c == '"' || c == '\'') {
 			url->st = DCC_URL_ST_IDLE;
 			RET_C(DCC_CK_URL_SKIP);
 		}
 		if ((c == '>' || c == ' ' || c == '\t'
 		    || c == '\n' || c == '\r')
-		    && (!(url->flags & DCC_URL_QUOTED)
+		    && (!(url->flags & DCC_URL_QUOTES)
 			|| url->total > DCC_URL_FAILSAFE)) {
 			url->total = 0;
 			url->st = DCC_URL_ST_IDLE;
